@@ -7,7 +7,9 @@ import os
 import random
 import binascii
 import requests
-from .errors import ValidationError,RequestError
+import time
+import shutil
+from .errors import ValidationError, RequestError
 from .crypto import *
 
 
@@ -15,14 +17,14 @@ class Mega(object):
     def __init__(self):
         self.schema = 'https'
         self.domain = 'mega.co.nz'
-        self.timeout = 160 #max time (secs) to wait for resp from api requests
+        self.timeout = 160  # max time (secs) to wait for resp from api requests
         self.sid = None
         self.sequence_num = random.randint(0, 0xFFFFFFFF)
         self.request_id = make_id(10)
 
     @classmethod
-    def login(class_, email, password):
-        instance = class_()
+    def login(cls, email, password):
+        instance = cls()
         instance.login_user(email, password)
         return instance
 
@@ -38,7 +40,6 @@ class Mega(object):
     def _login_process(self, resp, password):
         encrypted_master_key = base64_to_a32(resp['k'])
         self.master_key = decrypt_key(encrypted_master_key, password)
-        self.users_keys = dict()
         if 'tsid' in resp:
             tsid = base64_url_decode(resp['tsid'])
             key_encrypted = a32_to_str(
@@ -54,7 +55,7 @@ class Mega(object):
             self.rsa_private_key = [0, 0, 0, 0]
 
             for i in range(4):
-                l = ((ord(private_key[0])*256+ord(private_key[1]) +7) / 8) + 2
+                l = ((ord(private_key[0]) * 256 + ord(private_key[1]) + 7) / 8) + 2
                 self.rsa_private_key[i] = mpi_to_int(private_key[:l])
                 private_key = private_key[l:]
 
@@ -81,9 +82,9 @@ class Mega(object):
 
         req = requests.post(
             '{0}://g.api.{1}/cs'.format(self.schema, self.domain),
-                                        params=params,
-                                        data=json.dumps(data),
-                                        timeout=self.timeout)
+            params=params,
+            data=json.dumps(data),
+            timeout=self.timeout)
         json_resp = json.loads(req.text)
 
         #if numeric error code response
@@ -100,23 +101,33 @@ class Mega(object):
         else:
             raise RequestError('Url key missing')
 
-    def process_file(self, file):
+    def process_file(self, file, shared_keys):
         """
         Process a file
         """
         if file['t'] == 0 or file['t'] == 1:
+            keys = dict(keypart.split(':', 1) for keypart in file['k'].split('/'))
             uid = file['u']
-            keys = file['k'].split('/')
-            regex = re.compile('^%s:.*$' % uid)
             key = None
-            for keytmp in keys:
-                if regex.match(keytmp):
-                    key = keytmp[keytmp.index(':') + 1:]
-                    break
-
             # my objects
-            if key:
-                key = decrypt_key(base64_to_a32(key), self.master_key)
+            if uid in keys:
+                key = decrypt_key(base64_to_a32(keys[uid]), self.master_key)
+            # shared folders 
+            elif 'su' in file and 'sk' in file and ':' in file['k']:
+                shared_key = decrypt_key(base64_to_a32(file['sk']), self.master_key)
+                key = decrypt_key(base64_to_a32(keys[file['h']]), shared_key)
+                if file['su'] not in shared_keys:
+                    shared_keys[file['su']] = {}
+                shared_keys[file['su']][file['h']] = shared_key
+            # shared files
+            elif file['u'] and file['u'] in shared_keys:
+                for hkey in shared_keys[file['u']]:
+                    shared_key = shared_keys[file['u']][hkey]
+                    if hkey in keys:
+                        key = keys[hkey]
+                        key = decrypt_key(base64_to_a32(key), shared_key)
+                        break
+            if key is not None:
                 # file
                 if file['t'] == 0:
                     k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6],
@@ -126,39 +137,8 @@ class Mega(object):
                 # folder
                 else:
                     k = key
-                attributes = base64_url_decode(file['a'])
-                attributes = decrypt_attr(attributes, k)
-                file['a'] = attributes
-            # shared folders
-            elif 'su' in file and 'sk' in file and ':' in file['k']:
-                user_key = decrypt_key(base64_to_a32(file['sk']),
-                                       self.master_key)
-                key = decrypt_key(base64_to_a32(file['k'].split(':')[1]),
-                                                user_key)
-                # save user_key to decrypt shared files
-                self.users_keys[file['su']] = user_key
-                if file['t'] == 0:
-                    k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6],
-                         key[3] ^ key[7])
-                    file['iv'] = key[4:6] + (0, 0)
-                    file['meta_mac'] = key[6:8]
-                else:
-                    k = key
-                attributes = base64_url_decode(file['a'])
-                attributes = decrypt_attr(attributes, k)
-                file['a'] = attributes
-            # shared files
-            elif file['u'] and ':' in file['k']:
-                user_key = self.users_keys[file['u']]
-                key = decrypt_key(base64_to_a32(file['k'].split(':')[1]),
-                                  user_key)
-                if file['t'] == 0:
-                    k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6],
-                         key[3] ^ key[7])
-                    file['iv'] = key[4:6] + (0, 0)
-                    file['meta_mac'] = key[6:8]
-                else:
-                    k = key
+                file['key'] = key
+                file['k'] = k
                 attributes = base64_url_decode(file['a'])
                 attributes = decrypt_attr(attributes, k)
                 file['a'] = attributes
@@ -176,35 +156,55 @@ class Mega(object):
             file['a'] = {'n': 'Rubbish Bin'}
         return file
 
+    def init_shared_keys(self, files, shared_keys):
+        """
+        Init shared key not associated with a user.
+        Seems to happen when a folder is shared,
+        some files are exchanged and then the
+        folder is un-shared.
+        Keys are stored in files['s'] and files['ok']
+        """
+        ok_dict = {}
+        for ok_item in files['ok']:
+            shared_key = decrypt_key(base64_to_a32(ok_item['k']), self.master_key)
+            ok_dict[ok_item['h']] = shared_key
+        for s_item in files['s']:
+            if s_item['u'] not in shared_keys:
+                shared_keys[s_item['u']] = {}
+            if s_item['h'] in ok_dict:
+                shared_keys[s_item['u']][s_item['h']] = ok_dict[s_item['h']]
+
     ##########################################################################
     # GET
     def find(self, filename):
-        '''
+        """
         Return file object from given filename
-        '''
+        """
         files = self.get_files()
         for file in files.items():
             if file[1]['a'] and file[1]['a']['n'] == filename:
                 return file
 
     def get_files(self):
-        '''
+        """
         Get all files in account
-        '''
+        """
         files = self.api_request({'a': 'f', 'c': 1})
         files_dict = {}
+        shared_keys = {}
+        self.init_shared_keys(files, shared_keys)
         for file in files['f']:
-            processed_file = self.process_file(file)
+            processed_file = self.process_file(file, shared_keys)
             #ensure each file has a name before returning
             if processed_file['a']:
                 files_dict[file['h']] = processed_file
         return files_dict
 
     def get_upload_link(self, file):
-        '''
+        """
         Get a files public link inc. decrypted key
         Requires upload() response as input
-        '''
+        """
         if 'f' in file:
             file = file['f'][0]
             public_handle = self.api_request({'a': 'l', 'n': file['h']})
@@ -220,15 +220,15 @@ class Mega(object):
                             use get_link() for regular file input''')
 
     def get_link(self, file):
-        '''
+        """
         Get a file public link from given file object
-        '''
+        """
         file = file[1]
         if 'h' in file and 'k' in file:
             public_handle = self.api_request({'a': 'l', 'n': file['h']})
-            file_key = file['k'][file['k'].index(':') + 1:]
-            decrypted_key = a32_to_base64(decrypt_key(base64_to_a32(file_key),
-                                                      self.master_key))
+            if public_handle == -11:
+                raise RequestError("Can't get a public link from that file (is this a shared file?)")
+            decrypted_key = a32_to_base64(file['key'])
             return '{0}://{1}/#!{2}!{3}'.format(self.schema,
                                                 self.domain,
                                                 public_handle,
@@ -241,28 +241,30 @@ class Mega(object):
         return user_data
 
     def get_node_by_type(self, type):
-        '''
+        """
         Get a node by it's numeric type id, e.g:
         0: file
         1: dir
         2: special: root cloud drive
         3: special: inbox
         4: special trash bin
-        '''
+        """
         nodes = self.get_files()
         for node in nodes.items():
             if (node[1]['t'] == type):
                 return node
 
     def get_files_in_node(self, target):
-        '''
+        """
         Get all files in a given target, e.g. 4=trash
-        '''
+        """
         node_id = self.get_node_by_type(target)
         files = self.api_request({'a': 'f', 'c': 1})
         files_dict = {}
+        shared_keys = {}
+        self.init_shared_keys(files, shared_keys)
         for file in files['f']:
-            processed_file = self.process_file(file)
+            processed_file = self.process_file(file, shared_keys)
             if processed_file['a'] and processed_file['p'] == node_id[0]:
                 files_dict[file['h']] = processed_file
         return files_dict
@@ -280,18 +282,18 @@ class Mega(object):
         return node_id
 
     def get_quota(self):
-        '''
+        """
         Get current remaining disk quota in MegaBytes
-        '''
+        """
         json_resp = self.api_request({'a': 'uq', 'xfer': 1})
         #convert bytes to megabyes
-        return json_resp['mstrg']/1048576
+        return json_resp['mstrg'] / 1048576
 
     def get_balance(self):
-        '''
+        """
         Get account monetary balance, Pro accounts only
-        '''
-        user_data = self.api_request({"a":"uq","pro":1})
+        """
+        user_data = self.api_request({"a": "uq", "pro": 1})
         if 'balance' in user_data:
             return user_data['balance']
 
@@ -313,6 +315,7 @@ class Mega(object):
         return self.api_request({'a': 'd',
                                  'n': file_id,
                                  'i': self.request_id})
+
     def destroy_url(self, url):
         #delete a file via it's url
         path = self.parse_url(url).split('!')
@@ -321,7 +324,7 @@ class Mega(object):
         return self.destroy(file_id)
 
     def move(self, file_id, target):
-        '''
+        """
         Move a file to another parent node
         params:
         a : command
@@ -333,7 +336,7 @@ class Mega(object):
         2 : root
         3 : inbox
         4 : trash
-        '''
+        """
 
         #determine target_node_id
         target_node_id = str(self.get_node_by_type(target)[0])
@@ -342,7 +345,6 @@ class Mega(object):
                                  't': target_node_id,
                                  'i': self.request_id})
 
-
     def empty_trash(self):
         # get list of files in rubbish out
         files = self.get_files_in_node(4)
@@ -359,49 +361,56 @@ class Mega(object):
     ##########################################################################
     # DOWNLOAD
     def download(self, file, dest_path=None):
-        '''
+        """
         Download a file by it's file object
-        '''
-        url = self.get_link(file)
-        self.download_url(url, dest_path)
+        """
+        self.download_file(None, None, file=file[1], dest_path=dest_path, is_public=False)
 
     def download_url(self, url, dest_path=None):
-        '''
+        """
         Download a file by it's public url
-        '''
+        """
         path = self.parse_url(url).split('!')
         file_id = path[0]
         file_key = path[1]
         self.download_file(file_id, file_key, dest_path, is_public=True)
 
-    def download_file(self, file_handle, file_key, dest_path=None, is_public=False):
-        if is_public:
-            file_key = base64_to_a32(file_key)
-            file_data = self.api_request({'a': 'g', 'g': 1, 'p': file_handle})
+    def download_file(self, file_handle, file_key, dest_path=None, is_public=False, file=None):
+        if file is None :
+            if is_public:
+                file_key = base64_to_a32(file_key)
+                file_data = self.api_request({'a': 'g', 'g': 1, 'p': file_handle})
+            else:
+                file_data = self.api_request({'a': 'g', 'g': 1, 'n': file_handle})
+
+            k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
+                 file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
+            iv = file_key[4:6] + (0, 0)
+            meta_mac = file_key[6:8]
         else:
-            file_data = self.api_request({'a': 'g', 'g': 1, 'n': file_handle})
-
-        k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
-             file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
-        iv = file_key[4:6] + (0, 0)
-        meta_mac = file_key[6:8]
+            file_data = self.api_request({'a': 'g', 'g': 1, 'n': file['h']})
+            k = file['k']
+            iv = file['iv']
+            meta_mac = file['meta_mac']
 
+        # Seems to happens sometime... When  this occurs, files are 
+        # inaccessible also in the official also in the official webapp.
+        # Strangely, files can come back later.
+        if 'g' not in file_data:
+            raise RequestError('File not accessible anymore')
         file_url = file_data['g']
         file_size = file_data['s']
         attribs = base64_url_decode(file_data['at'])
         attribs = decrypt_attr(attribs, k)
         file_name = attribs['n']
-
-        print("downloading {0} (size: {1}), url = {2}".format(attribs['n'].encode("utf8"),
-                                                              file_size,
-                                                              file_url))
+        file_name_tmp = '.megapy-%s-%s' % (int(time.time() * 1000), file_name)
 
         input_file = requests.get(file_url, stream=True).raw
 
         if dest_path:
-            output_file = open(dest_path + '/' + file_name, 'wb')
+            output_file = open(dest_path + '/' + file_name_tmp, 'wb')
         else:
-            output_file = open(file_name, 'wb')
+            output_file = open(file_name_tmp, 'wb')
 
         counter = Counter.new(
             128, initial_value=((iv[0] << 32) + iv[1]) << 64)
@@ -438,6 +447,8 @@ class Mega(object):
         if (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]) != meta_mac:
             raise ValueError('Mismatched mac')
 
+        shutil.move(file_name_tmp, file_name)
+
     ##########################################################################
     # UPLOAD
     def upload(self, filename, dest=None):
@@ -454,8 +465,8 @@ class Mega(object):
         ul_url = self.api_request({'a': 'u', 's': size})['p']
 
         #generate random aes key (128) for file
-        ul_key = [random.randint(0, 0xFFFFFFFF) for r in range(6)]
-        count = Counter.new(128,initial_value=((ul_key[4]<<32)+ul_key[5])<<64)
+        ul_key = [random.randint(0, 0xFFFFFFFF) for _ in range(6)]
+        count = Counter.new(128, initial_value=((ul_key[4] << 32) + ul_key[5]) << 64)
         aes = AES.new(a32_to_str(ul_key[:4]), AES.MODE_CTR, counter=count)
 
         file_mac = [0, 0, 0, 0]
@@ -514,7 +525,7 @@ class Mega(object):
             dest = self.root_id
 
         #generate random aes key (128) for folder
-        ul_key = [random.randint(0, 0xFFFFFFFF) for r in range(6)]
+        ul_key = [random.randint(0, 0xFFFFFFFF) for _ in range(6)]
 
         #encrypt attribs
         attribs = {'n': name}
